#!perl
# Copyright (c) 2003-2009 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

# See comment at top of rxx_python.py for a description of the protocol

use strict;
use warnings;

use JSON 2.0 qw(from_json to_json);
## no critic (ProhibitStringyEval)

package Evaluator;

sub new {
    my ($class, $requestString) = @_;
    my $requestPacket = JSON::from_json($requestString);
    my $hash = {
        op => $requestPacket->{operation},
        pattern => $requestPacket->{pattern},
        options => $requestPacket->{options},
        subjectText => $requestPacket->{text},
        requestPacket => $requestPacket,
    };
    my $self = bless $hash, $class;
    return $self;
}

sub init {
    my $self = shift;
    $self->{regex} = $self->_compile();
}

sub _compile {
    # Convert post-pattern options to pre-pattern options like
    # (?i)(?x)...
    # and then use qr// to eval the whole shebang.  Mmm, Perl.
    my $self = shift;
    my $optionsPrefix = $self->{options};
    $optionsPrefix =~ s/(.)/(?$1)/g;
    my $pattern = $optionsPrefix . $self->{pattern};
    return qr{$pattern};
}

sub _groups_from_match_obj {
    my ($self, $matchStartsA, $matchEndsA, $substitution) = @_;
    my $group = {name => undef,
                 span => [$matchStartsA->[0], $matchEndsA->[0]],
                 value => substr($self->{subjectText},
                                 $matchStartsA->[0],
                                 $matchEndsA->[0] - $matchStartsA->[0]),
                 };
    if (defined $substitution) {
        $group->{replacement} = $substitution;
    }
    my $groups = [$group];
    for my $i (1 .. $#$matchStartsA) {
        $group = {
            name => undef,
            span => [$matchStartsA->[$i], $matchEndsA->[$i]],
            value => substr($self->{subjectText},
                            $matchStartsA->[$i],
                            $matchEndsA->[$i] - $matchStartsA->[$i]),
        };
        if (defined $substitution) {
            $group->{replacement} = "";
        }
        push @$groups, $group;
    }
    return $groups;
}

sub do_match {
    my $self = shift;
    if ($self->{subjectText} =~ $self->{regex}) {
        my $groups = $self->_groups_from_match_obj(\@-, \@+);
        return { 'status' => 'ok',
                     'result' => [$groups],
                     'lastNumGroups' => 0+@- - 1}
    } else {
        return { 'status' => 'matchFailure' }
    }
}

sub do_matchAll {
    my $self = shift;
    my $groupObjs = [];
    my $start = 0;
    while ($self->{subjectText} =~ /$self->{regex}/g) {
        push @$groupObjs, $self->_groups_from_match_obj(\@-, \@+);
    }
    if (0+ @$groupObjs == 0) {
        return { 'status' => 'matchFailure' }
    };
    my $lastGroup = $groupObjs->[$#$groupObjs];
    return { 'status' => 'ok',
            'result' => $groupObjs,
            'lastNumGroups' => $#$lastGroup,
           }
}

sub do_split {
    my $self = shift;
    my $res = $self->do_match();
    if ($res->{status} ne 'ok') {
        return $res;
    }
    $res->{result} = [split($self->{regex}, $self->{subjectText})];
    return $res;
}

sub do_replace {
    my $self = shift;
    my $res = $self->do_match();
    if ($res->{status} ne 'ok') {
        return $res;
    }
    my $replacedText = $self->{subjectText};
    my $replacement = $self->{requestPacket}{replacement};
    my $ptn = $self->{regex};
    # I can't get this to work using the s///e form:
    #$replacedText =~ s/$self->{regex}/$replacement/e;
    my $expn = "\$replacedText =~ s{\$ptn}{$replacement}";
    eval($expn);
    if ($@) {
        return {
            status => 'matchError',
            exception => $@,
        };
    }
    $res->{replacedText} = $replacedText;
    my @substitutions;
    # Now figure out the substitution, using the same technique as in replaceAll
    my ($text, $originalText);
    my $processThisMatch = sub {
        my ($matchedPart) = @_;
        my $matchedPartStart = index($originalText, $matchedPart);
        my $diff = length($text) - length($originalText);
        my $thisSub = substr($text, $matchedPartStart,
                             length($matchedPart) + $diff);
        push @substitutions, $thisSub;
    };
    $originalText = $text = $self->{subjectText};
    eval qq{
        if (\$text =~ s/\$ptn/$replacement/) {
            \&\$processThisMatch(\$\&);
        }
    };
    if ($@) {
        return {
            status => 'matchError',
            exception => $@,
        };
    }
    $res->{substitutions} = \@substitutions;
    return $res;
}

sub do_replaceAll {
    my $self = shift;
    my $res = $self->do_matchAll();
    if ($res->{status} ne 'ok') {
        return $res;
    }
    my $replacedText = $self->{subjectText};
    my $replacement = $self->{requestPacket}{replacement};
    my $expn = "\$replacedText =~ s{\$self->{regex}}{$replacement}g";
    eval($expn);
    $res->{replacedText} = $replacedText;

    # This part's tricky, because Perl doesn't make it easy to
    # get the list of replaced parts.  We have to rerun the
    # sub in a while loop in an eval
    
    my @substitutions = ();
    my ($inputPos, $runningDiff);
    my ($text, $originalText, $beforeText);
    
    my $processThisMatch = sub {
        my ($matchedPart) = @_;
        my $matchedPartStart = $inputPos + index(substr($originalText, $inputPos), $matchedPart);
        $inputPos = $matchedPartStart + length($matchedPart);
        my $diff = length($text) - length($beforeText);
        my $thisSub = substr($text, $matchedPartStart + $runningDiff,
                             length($matchedPart) + $diff);
        push @substitutions, $thisSub;
        $runningDiff += $diff;
        $beforeText = $text;
    };
    $originalText = $beforeText = $text = $self->{subjectText};
    $inputPos = $runningDiff = 0;
    my $ptn = $self->{regex};
    eval qq{
        while (\$text =~ s/\$ptn/$replacement/) {
            \&\$processThisMatch(\$\&);
        }
    };
    if ($@) {
        return {
            status => 'matchError',
            exception => $@,
        };
    }
    $res->{substitutions} = \@substitutions;
    return $res;
}

sub run {
    my $self = shift;
    my $funcName = "do_" . $self->{op};
    my $res = $self->$funcName();
    $res->{operation} = $self->{op};
    $res->{lastGroupNames} = []; #Not available in Perl
    return $res;
}

package main;

sub main {
    my $requestString = shift || <STDIN>;
    local $@;
    my $evaluator;
    eval {
        $evaluator = new Evaluator($requestString);
        $evaluator->init();
    };
    if ($@) {
        if ($@ =~ /^(.*?) in regex; marked by <-- HERE in/) {
            return {
                status => 'matchError',
                exception => $1,
            };
        } elsif ($@ =~ /^(.*?) in regex .*? at .*rxx_perl.pl line/) {
            return {
                status => 'matchError',
                exception => $1,
            };
        } else {
            return {
                status => 'matchError',
                exception => $@,
            };
        }
    }
    return $evaluator->run();
}

unless (caller) {
    my $str = "";
    my $packet = {
        text => 'cole',
        pattern => '(\w)',
        operation => 'matchAll',
        replacement => '<<[$2]v[$1]>>',
        options => '',
    };
    #$str = JSON::to_json($packet);
    my $responsePacket = main($str);
    my $jsonResult = JSON::to_json($responsePacket, { ascii => 1 });
    print $jsonResult;
}
